@@ -43,6 +43,7 @@
 #endif
 
 using namespace std;
+using base::Time;
 using namespace iodrivers_base;
 
 string Driver::printable_com(std::string const& str)
@@ -87,7 +88,7 @@ string Driver::binary_com(char const* str, size_t str_size)
 Driver::Driver(int max_packet_size, bool extract_last)
     : internal_buffer(new uint8_t[max_packet_size]), internal_buffer_size(0)
     , MAX_PACKET_SIZE(max_packet_size)
-    , m_stream(0), m_auto_close(true), m_extract_last(extract_last) 
+    , m_stream(0), m_auto_close(true), m_extract_last(extract_last)
 {
     if(MAX_PACKET_SIZE <= 0)
         std::runtime_error("Driver: max_packet_size cannot be smaller or equal to 0!");
@@ -140,9 +141,9 @@ void Driver::resetStatus()
 void Driver::setExtractLastPacket(bool flag) { m_extract_last = flag; }
 bool Driver::getExtractLastPacket() const { return m_extract_last; }
 
-void Driver::setFileDescriptor(int fd, bool auto_close)
+void Driver::setFileDescriptor(int fd, bool auto_close, bool has_eof)
 {
-    setMainStream(new FDStream(fd, auto_close));
+    setMainStream(new FDStream(fd, auto_close, has_eof));
 }
 
 int Driver::getFileDescriptor() const
@@ -237,7 +238,7 @@ void Driver::openTestMode()
 
 bool Driver::openSerial(std::string const& port, int baud_rate)
 {
-    setFileDescriptor(Driver::openSerialIO(port, baud_rate));
+    setFileDescriptor(Driver::openSerialIO(port, baud_rate), true, false);
     return true;
 }
 
@@ -307,7 +308,7 @@ static int createIPClientSocket(const char *hostname, const char *port, addrinfo
     if (addr_len != NULL) *addr_len = rp->ai_addrlen;
 
     freeaddrinfo(result);
-    
+
     return sfd;
 }
 
@@ -345,7 +346,7 @@ void Driver::openUDP(std::string const& hostname, int port)
         hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
         hints.ai_socktype = SOCK_DGRAM; /* Datagram socket */
         hints.ai_flags = AI_PASSIVE;    /* For wildcard IP address */
-        
+
         int sfd = createIPServerSocket(port, hints);
         setMainStream(new UDPServerStream(sfd,true));
     }
@@ -375,7 +376,7 @@ void Driver::openUDPBidirectional(std::string const& hostname, int out_port, int
     struct sockaddr peer;
     size_t peer_len;
     int peerfd = createIPClientSocket(hostname.c_str(), boost::lexical_cast<string>(out_port).c_str(), out_hints, &peer, &peer_len);
-    ::close(peerfd);    
+    ::close(peerfd);
 
     int sfd = createIPServerSocket(in_port, in_hints);
     setMainStream(new UDPServerStream(sfd, true, &peer, &peer_len));
@@ -479,9 +480,9 @@ bool Driver::setSerialBaudrate(int fd, int brate) {
 	ss.custom_divisor = (ss.baud_base + (brate / 2)) / brate;
 	int closestSpeed = ss.baud_base / ss.custom_divisor;
 
-	if (closestSpeed < brate * 98 / 100 || closestSpeed > brate * 102 / 100) 
+	if (closestSpeed < brate * 98 / 100 || closestSpeed > brate * 102 / 100)
 	{
-	    std::cerr << "Cannot set custom serial rate to " << brate 
+	    std::cerr << "Cannot set custom serial rate to " << brate
 		<< ". The closest possible value is " << closestSpeed << "."
 		<< std::endl;
 	}
@@ -495,22 +496,22 @@ bool Driver::setSerialBaudrate(int fd, int brate) {
 
     struct termios termios_p;
     if(tcgetattr(fd, &termios_p)){
-        perror("Failed to get terminal info \n");    
+        perror("Failed to get terminal info \n");
         return false;
     }
 
     if(cfsetispeed(&termios_p, tc_rate)){
-        perror("Failed to set terminal input speed \n");    
+        perror("Failed to set terminal input speed \n");
         return false;
     }
 
     if(cfsetospeed(&termios_p, tc_rate)){
-        perror("Failed to set terminal output speed \n");    
+        perror("Failed to set terminal output speed \n");
         return false;
     }
 
     if(tcsetattr(fd, TCSANOW, &termios_p)) {
-        perror("Failed to set speed \n");    
+        perror("Failed to set speed \n");
         return false;
     }
     return true;
@@ -526,13 +527,13 @@ std::pair<uint8_t const*, int> Driver::findPacket(uint8_t const* buffer, int buf
 {
     int packet_start = 0, packet_size = 0;
     int extract_result = extractPacket(buffer, buffer_size);
-    
+
     // make sure the returned packet size is not longer than
     // the buffer
     if( extract_result > buffer_size )
         throw length_error("extractPacket() returned result size "
-                + boost::lexical_cast<string>(extract_result) 
-                + ", which is larger than the buffer size " 
+                + boost::lexical_cast<string>(extract_result)
+                + ", which is larger than the buffer size "
                 + boost::lexical_cast<string>(buffer_size) + ".");
 
     if (0 == extract_result)
@@ -545,7 +546,7 @@ std::pair<uint8_t const*, int> Driver::findPacket(uint8_t const* buffer, int buf
 
     if (m_extract_last)
     {
-        m_stats.stamp = base::Time::now();
+        m_stats.stamp = Time::now();
         m_stats.bad_rx  += packet_start;
         m_stats.good_rx += packet_size;
     }
@@ -585,17 +586,12 @@ int Driver::doPacketExtraction(uint8_t* buffer)
     pair<uint8_t const*, int> packet = findPacket(internal_buffer, internal_buffer_size);
     if (!m_extract_last)
     {
-        m_stats.stamp = base::Time::now();
+        m_stats.stamp = Time::now();
         m_stats.bad_rx  += packet.first - internal_buffer;
         m_stats.good_rx += packet.second;
     }
-    // cerr << "found packet " << printable_com(packet.first, packet.second) << " in internal buffer" << endl;
-
-    int buffer_rem = internal_buffer_size - (packet.first + packet.second - internal_buffer);
-    memcpy(buffer, packet.first, packet.second);
-    memmove(internal_buffer, packet.first + packet.second, buffer_rem);
-    internal_buffer_size = buffer_rem;
 
+    pullBytesFromInternal(buffer, packet.first - internal_buffer, packet.second);
     return packet.second;
 }
 
@@ -619,6 +615,74 @@ pair<int, bool> Driver::extractPacketFromInternalBuffer(uint8_t* buffer, int out
     return make_pair(result_size, false);
 }
 
+void Driver::pullBytesFromInternal(uint8_t* buffer, int skip, int size) {
+    int total_size = skip + size;
+    int new_internal_size = internal_buffer_size - total_size;
+
+    memcpy(buffer, internal_buffer + skip, size);
+    memmove(internal_buffer,
+            internal_buffer + total_size,
+            new_internal_size);
+    internal_buffer_size = new_internal_size;
+}
+
+int Driver::readRaw(uint8_t* buffer, int out_buffer_size)
+{
+    return readRaw(buffer, out_buffer_size, getReadTimeout());
+}
+
+int Driver::readRaw(uint8_t* buffer, int out_buffer_size, Time const& timeout)
+{
+    return readRaw(buffer, out_buffer_size, timeout, timeout);
+}
+
+int Driver::readRaw(uint8_t* buffer, int out_buffer_size,
+                    Time const& packet_timeout,
+                    Time const& first_byte_timeout_,
+                    Time const& inter_byte_timeout_) {
+    if (!isValid()) {
+        throw std::runtime_error("attempting to call readRaw on a closed driver");
+    }
+
+    int buffer_fill = std::min<int>(internal_buffer_size, out_buffer_size);
+    pullBytesFromInternal(buffer, 0, buffer_fill);
+
+    auto first_byte_timeout = min(packet_timeout, first_byte_timeout_);
+    auto inter_byte_timeout = inter_byte_timeout_.isNull() ?
+                              packet_timeout : inter_byte_timeout_;
+
+    auto now = Time::now();
+    auto last_char = now + packet_timeout;
+    bool received_bytes = false;
+    Time global_deadline = now + first_byte_timeout;
+    while (buffer_fill < out_buffer_size && now <= global_deadline)
+    {
+        auto deadline = min(global_deadline, last_char + inter_byte_timeout);
+        try {
+            m_stream->waitRead(deadline - now);
+        }
+        catch(TimeoutError&) {
+            break;
+        }
+        int c = m_stream->read(buffer + buffer_fill,
+                               out_buffer_size - buffer_fill);
+        now = Time::now();
+
+        if (c > 0) {
+            last_char = now;
+            if (!received_bytes) {
+                global_deadline = now + packet_timeout;
+                received_bytes = true;
+            }
+            for (IOListener* it: m_listeners)
+                it->readData(buffer + buffer_fill, c);
+        }
+        buffer_fill += c;
+    }
+
+    return buffer_fill;
+}
+
 pair<int, bool> Driver::readPacketInternal(uint8_t* buffer, int out_buffer_size)
 {
     if (out_buffer_size < MAX_PACKET_SIZE)
@@ -642,7 +706,7 @@ pair<int, bool> Driver::readPacketInternal(uint8_t* buffer, int out_buffer_size)
         if (c > 0) {
             for (set<IOListener*>::iterator it = m_listeners.begin(); it != m_listeners.end(); ++it)
                 (*it)->readData(internal_buffer + internal_buffer_size, c);
-	  
+
             received_something = true;
 
             // cerr << "received: " << printable_com(buffer + buffer_size, c) << endl;
@@ -676,122 +740,124 @@ bool Driver::hasPacket() const
     return (packet.second > 0);
 }
 
-void Driver::setReadTimeout(base::Time const& timeout)
+void Driver::setReadTimeout(Time const& timeout)
 { m_read_timeout = timeout; }
-base::Time Driver::getReadTimeout() const
+Time Driver::getReadTimeout() const
 { return m_read_timeout; }
 int Driver::readPacket(uint8_t* buffer, int buffer_size)
 {
-    return readPacket(buffer, buffer_size, getReadTimeout());
+    return readPacket(buffer, buffer_size,
+                      getReadTimeout(), getReadTimeout());
+}
+int Driver::readPacket(uint8_t* buffer, int buffer_size, Time const& packet_timeout)
+{
+    return readPacket(buffer, buffer_size,
+                      packet_timeout, packet_timeout);
 }
 int Driver::readPacket(uint8_t* buffer, int buffer_size,
-        base::Time const& packet_timeout)
+                       int packet_timeout, int first_byte_timeout)
 {
-    return readPacket(buffer, buffer_size, packet_timeout,
-            packet_timeout + base::Time::fromSeconds(1));
-}
-int Driver::readPacket(uint8_t* buffer, int buffer_size,
-        base::Time const& packet_timeout, base::Time const& first_byte_timeout)
-{
-    return readPacket(buffer, buffer_size, packet_timeout.toMilliseconds(), 
-            first_byte_timeout.toMilliseconds());
-}
-int Driver::readPacket(uint8_t* buffer, int buffer_size, int packet_timeout, int first_byte_timeout)
-{
-    if (first_byte_timeout > packet_timeout)
-        first_byte_timeout = -1;
+    if (first_byte_timeout == -1) {
+        first_byte_timeout = packet_timeout;
+    }
 
-    if (buffer_size < MAX_PACKET_SIZE)
+    return readPacket(buffer, buffer_size,
+                      Time::fromMilliseconds(packet_timeout),
+                      Time::fromMilliseconds(first_byte_timeout));
+}
+int Driver::readPacket(uint8_t* buffer, int buffer_size,
+                       Time const& packet_timeout, Time const& first_byte_timeout_)
+{
+    if (buffer_size < MAX_PACKET_SIZE) {
         throw length_error("readPacket(): provided buffer too small (got "
                 + boost::lexical_cast<string>(buffer_size) + ", expected at least "
                 + boost::lexical_cast<string>(MAX_PACKET_SIZE) + ")");
+    }
 
-    if (!isValid())
-    {
+    if (!isValid()) {
         // No valid file descriptor. Assume that the user is using the raw data
         // interface (i.e. that the data is already in the internal read buffer)
         pair<int, bool> result = extractPacketFromInternalBuffer(buffer, buffer_size);
-        if (result.first)
+        if (result.first) {
             return result.first;
-        else
-            throw TimeoutError(TimeoutError::PACKET,
-                    "readPacket(): no packet in the internal buffer and no FD to read from");
+        }
+        else {
+            throw TimeoutError(
+                TimeoutError::PACKET, "readPacket(): no packet in the internal "\
+                                      "buffer and no stream to read from");
+        }
     }
 
-    if(!m_stream)
-        throw std::runtime_error("Driver::writePacket : invalid stream, did you forget to call open ?");
+    TimeoutError::TIMEOUT_TYPE timeout_type = TimeoutError::FIRST_BYTE;
+    Time first_byte_timeout = min(packet_timeout, first_byte_timeout_);
+    Time start_time = Time::now();
+    Time deadline = start_time + first_byte_timeout;
 
-    Timeout time_out;
-    bool read_something = false;
-    while(true) {
-	
+    while (true) {
         pair<int, bool> read_state = readPacketInternal(buffer, buffer_size);
-            
         int packet_size = read_state.first;
-            
-        read_something = read_something || read_state.second;
-        
-        if (packet_size > 0)
+
+        if (packet_size > 0) {
             return packet_size;
+        }
 
         // if there was no data to read _and_ packet_timeout is zero, we'll throw
-        if (packet_timeout == 0)
-            throw TimeoutError(TimeoutError::FIRST_BYTE,
-                    "readPacket(): no data to read while a packet_timeout of 0 was given");
-
-        int timeout;
-        TimeoutError::TIMEOUT_TYPE timeout_type;
-        if (first_byte_timeout != -1 && !read_something)
-        {
-            timeout = first_byte_timeout;
-            timeout_type = TimeoutError::FIRST_BYTE;
+        if (packet_timeout.isNull()) {
+            throw TimeoutError(
+                TimeoutError::FIRST_BYTE,
+                "readPacket(): no data to read while a packet_timeout of 0 was given");
         }
-        else
-        {
-            timeout = packet_timeout;
+
+        bool read_something = read_state.second;
+        if (timeout_type == TimeoutError::FIRST_BYTE && read_something) {
+            deadline = start_time + packet_timeout;
             timeout_type = TimeoutError::PACKET;
         }
 
-        if (time_out.elapsed(timeout))
+        Time now = Time::now();
+        if (now > deadline)
         {
-            throw TimeoutError(timeout_type,
+            throw TimeoutError(
+                timeout_type,
                 "readPacket(): no data after waiting "
-                + boost::lexical_cast<string>(timeout) + "ms");
+                + boost::lexical_cast<string>((now - start_time).toMilliseconds()) + "ms");
         }
 
         // we still have time left to wait for arriving data. see how much
-        int remaining_timeout = time_out.timeLeft(timeout);
+        Time remaining = deadline - now;
         try {
             // calls select and waits until a new read can be actually performed (in the next
             // while-iteration)
-            m_stream->waitRead(base::Time::fromMicroseconds(remaining_timeout * 1000));
+            m_stream->waitRead(remaining);
         }
-        catch(TimeoutError& e)
+        catch (TimeoutError& e)
         {
+            auto total_wait = Time::now() - start_time;
             throw TimeoutError(timeout_type,
-                "readPacket(): no data after retrying with remaining time "
-                + boost::lexical_cast<string>(remaining_timeout) + "ms of "
-                + boost::lexical_cast<string>(timeout) +"ms timeout");
+                "readPacket(): no data waiting for data. Last wait lasted "
+                + boost::lexical_cast<string>(remaining.toMilliseconds()) + "ms, "
+                + "out of a total wait of " +
+                boost::lexical_cast<string>(total_wait.toMilliseconds()) + "ms");
         }
     }
 }
 
-void Driver::setWriteTimeout(base::Time const& timeout)
+void Driver::setWriteTimeout(Time const& timeout)
 { m_write_timeout = timeout; }
-base::Time Driver::getWriteTimeout() const
+Time Driver::getWriteTimeout() const
 { return m_write_timeout; }
 
 bool Driver::writePacket(uint8_t const* buffer, int buffer_size)
 {
     return writePacket(buffer, buffer_size, getWriteTimeout());
 }
-bool Driver::writePacket(uint8_t const* buffer, int buffer_size, base::Time const& timeout)
+bool Driver::writePacket(uint8_t const* buffer, int buffer_size, Time const& timeout)
 { return writePacket(buffer, buffer_size, timeout.toMilliseconds()); }
 bool Driver::writePacket(uint8_t const* buffer, int buffer_size, int timeout)
 {
     if(!m_stream)
         throw std::runtime_error("Driver::writePacket : invalid stream, did you forget to call open ?");
-    
+
     Timeout time_out(timeout);
     int written = 0;
     while(true) {
@@ -801,16 +867,22 @@ bool Driver::writePacket(uint8_t const* buffer, int buffer_size, int timeout)
         written += c;
 
         if (written == buffer_size) {
-            m_stats.stamp = base::Time::now();
-	    m_stats.tx += buffer_size;
+            m_stats.stamp = Time::now();
+            m_stats.tx += buffer_size;
             return true;
         }
-        
+
         if (time_out.elapsed())
             throw TimeoutError(TimeoutError::PACKET, "writePacket(): timeout");
 
         int remaining_timeout = time_out.timeLeft();
-        m_stream->waitWrite(base::Time::fromMicroseconds(remaining_timeout * 1000));
+        m_stream->waitWrite(Time::fromMicroseconds(remaining_timeout * 1000));
     }
 }
 
+bool Driver::eof() const
+{
+    if (!m_stream)
+        throw std::runtime_error("eof(): invalid stream");
+    return m_stream->eof();
+}
\ No newline at end of file
